/* Copyright (C) 2005-2006  Emiliano Betti and Marco Cesati
   Copyright (C) 2003-2004  Marco Cesati
   Copyright (C) 2002-2003  Shell Technologies s.r.l.

   Authors:
            Emiliano Betti <emilianobetti@gmail.com>
            Marco Cesati   <marcocesati@gmail.com>

   Driver for the Galil 1800 PCI motion controllers (http://www.galilmc.com/)

   This is a driver for the Linux kernel 2.6. It is based on an older driver
   for the kernel 2.4, which in turn was loosely inspired by Galil's own
   Linux driver, as well as by a bunch of PCI drivers in the kernel source code

   This program is free software; you can redistribute  it and/or modify it
   under  the terms of  the GNU General  Public License as published by the
   Free Software Foundation.
*/

#include <asm/uaccess.h>
#include <linux/fs.h>
#include <linux/poll.h>

#include "galil1800_io.h"
#include "galil1800_init.h"
#include "galil1800_sysfs.h"

/* Stop write on hardware write fifo.
   To make that it's performed an unsafe clean of write kernel buffer.
   This command is only for use in dangerous situation and after
   it's safe to reload the module!
*/
static void galil1800_unsafe_write_stop(
		struct galil1800_card_wbuffer *wbuff_p)
{
	struct galil1800_state *s = container_of(wbuff_p,
			struct galil1800_state, card_wbuff);
	if(wbuff_p->size) {
		memset(wbuff_p->flag, 0,
			GALIL1800_CARD_WB_FLAG_BITMAP_SZ(wbuff_p));
		wbuff_p->tail=wbuff_p->head=wbuff_p->data;
		wbuff_p->pause = 0UL;
	}
	galil1800_reset_fifo_rw(s);
	schedule_timeout_uninterruptible(HZ/100); /* sleep 10 msec. */
	galil1800_reset_fifo_rw(s);
	printk(KERN_ALERT PFX "Unsafe write stop performed!"
				" Please reload the module\n");
}


/* Check if the value for write buffer size is correct, and eventually
   correct it. Return 0 if no correction is needed, positive value othewise */
static int galil1800_card_wbuff_size_check(unsigned int *wbuff_size)
{
	unsigned int mod;
	if(!(*wbuff_size)) {
		printk(KERN_INFO PFX "Write buffer for card device disable\n");
		return 0;
	}
	if(*wbuff_size < GALIL1800_CARD_WBUFF_SZ_MIN) {
		*wbuff_size = GALIL1800_CARD_WBUFF_DEFAULT_SZ;
		printk(KERN_NOTICE PFX "Write buffer size for card device is "
			"less then %d, set default value (%d)\n",
			GALIL1800_CARD_WBUFF_SZ_MIN,
			GALIL1800_CARD_WBUFF_DEFAULT_SZ);
		return 1;
	}
	if(*wbuff_size > GALIL1800_CARD_WBUFF_SZ_MAX) {
		*wbuff_size = GALIL1800_CARD_WBUFF_DEFAULT_SZ;
		printk(KERN_NOTICE PFX "Write buffer size for card device is "
			"more then %d, set default value (%d)\n",
			GALIL1800_CARD_WBUFF_SZ_MAX,
			GALIL1800_CARD_WBUFF_DEFAULT_SZ);
		return 2;
	}
	mod = *wbuff_size %
		((sizeof(unsigned long)*8)/GALIL1800_CARD_WB_FLAG_BIT_FOR_BYTE);
	if(mod) {
		*wbuff_size -= mod;
		printk(KERN_NOTICE PFX "Write buffer size for card device must"
			" be multiple of %d, so truncated to (%d)\n",
			((sizeof(unsigned long)*8) /
			GALIL1800_CARD_WB_FLAG_BIT_FOR_BYTE), *wbuff_size);
		return 3;
	}
	return 0;
}

/* Check parameter for card device and eventually correct it
   Return 0 if no correction is needed, positive value othewise */
int galil1800_card_parameter_check(unsigned int *pause, unsigned int *wbuff_size)
{
	int ret;
	ret = galil1800_card_wbuff_size_check(wbuff_size);

	if(*pause < 0 || *pause > 1) {
		*pause = GALIL1800_CARD_PAUSE_ENABLE_DEFAULT;
		printk(KERN_WARNING PFX
			"pause_enalbe must be 1 (on) or 0 (off)."
			" Set default (%d)\n",
			GALIL1800_CARD_PAUSE_ENABLE_DEFAULT);
		ret++;
	}

	if(*pause && !(*wbuff_size)) {
		printk(KERN_WARNING PFX
			"To enable pseudo \"pause\" command must be"
			" enable the card write buffer!\n");
		*pause = 0;
		ret++;
	}

	return ret;
}


/* Second step of initialization: must be call in open device function */
static int init_galil1800_card_wbuffer_open_dev(
			struct galil1800_card_wbuffer *wbuff_p)
{
	int rc;
	char *wq_name;

	if(!(wbuff_p->size)) {
		GALIL_PRINT_DEBUG("Write Buffer disable\n");
		return 0;
	}

	GALIL_PRINT_DEBUG("Write Buffer enable\n");

	wq_name = kzalloc(strlen(GALIL1800_MODULE_NAME)+5, GFP_KERNEL);
	if (!wq_name) {
		printk(KERN_ERR PFX "init_galil1800_card_wbuffer_open_dev:"
						" alloc wq_name failed\n");
		rc = -ENOMEM;
		goto out0;
	}
	sprintf(wq_name, "%s_cwb", GALIL1800_MODULE_NAME);

	wbuff_p->wq = create_singlethread_workqueue(wq_name);
	kfree(wq_name);
	if (!(wbuff_p->wq)) {
		printk(KERN_ERR PFX "cannot create a workqueue "
					"for write card buffer\n");
		rc = -ENOMEM;
		goto out0;
	}

	wbuff_p->data = kzalloc(wbuff_p->size, GFP_KERNEL);
	if (!(wbuff_p->data)) {
		printk(KERN_ERR PFX "init_galil1800_card_wbuffer_open_dev:"
						" alloc data failed\n");
		rc = -ENOMEM;
		goto out1;
	}

	GALIL_PRINT_DEBUG("Card Write Buffer address %lx\n",
					(unsigned long)wbuff_p->data);

	wbuff_p->head = wbuff_p->tail = wbuff_p->data;

	wbuff_p->flag = kzalloc(GALIL1800_CARD_WB_FLAG_BITMAP_SZ(wbuff_p),
								GFP_KERNEL);
	if (!(wbuff_p->flag)) {
		printk(KERN_ERR PFX "init_galil1800_card_wbuffer_open_dev: "
							"alloc flag failed\n");
		rc = -ENOMEM;
		goto out2;
	}

	wbuff_p->pause = 0UL;

	wbuff_p->run = 1;
	if(!queue_work(wbuff_p->wq, &wbuff_p->work)) {
		printk(KERN_ERR PFX "init_galil1800_card_wbuffer_open_dev: "
							"queue_work failed\n");
		rc = -1;
		goto out3;
	}

	return 0;
out3:
	wbuff_p->run = 0;
	kfree(wbuff_p->flag);
out2:
	kfree(wbuff_p->data);
	wbuff_p->head = wbuff_p->tail = wbuff_p->data = 0;
out1:
	destroy_workqueue(wbuff_p->wq);
out0:
	return rc;
}

/* Clean for second step of initialization:
   must be call in release device function */
static
inline void clean_galil1800_card_wbuffer_close_dev(
			struct galil1800_card_wbuffer *wbuff_p)
{
	if(!(wbuff_p->size))
		return;

	wbuff_p->run = 0; /* to stop workqueue function */
	wake_up(&wbuff_p->read_wait);
	/* Next function block until the buffer will be empty */
	flush_workqueue(wbuff_p->wq);
	destroy_workqueue(wbuff_p->wq);
	kfree(wbuff_p->flag);
	kfree(wbuff_p->data);
	GALIL_PRINT_DEBUG("Card Write Buffer address %lx\n",
				(unsigned long)wbuff_p->data);
	wbuff_p->data = 0;
	wbuff_p->flag = 0;
	wbuff_p->head = 0;
	wbuff_p->tail = 0;
	wbuff_p->last_size_p10 = 0;
	wbuff_p->last_event_opt_flag = 0;
	wbuff_p->pause = 0UL;
}

/* Try to add len byte from data to buffer.
   Block if there is no space in buffer and nonblock==0.
   Return number of bytes added or
   -EAGAIN if nonblock is set and can't write or
   -ERESTARTSYS if signal is raised */
static int galil1800_card_wbuffer_add(struct galil1800_card_wbuffer *wbuff_p,
				unsigned char *data, size_t len, int nonblock)
{
	int rc = 0;

	/* Wait to buffer is not full */
	while(GALIL1800_CARD_WB_FULL(wbuff_p)) {
		if(nonblock)
			return -EAGAIN;
		wake_up(&wbuff_p->read_wait);
		GALIL_PRINT_DEBUG( "galil1800_card_wbuffer_add "
				"going to sleep (buffer full)\n");
		if(wait_event_interruptible(wbuff_p->write_wait,
					!GALIL1800_CARD_WB_FULL(wbuff_p))) {
			GALIL_PRINT_DEBUG_PID("\"card galil1800_card_"
				"wbuffer_add\": \"down interrupted\n");
			return -ERESTARTSYS;
		}
	}

	while(rc < len) {
		*(wbuff_p->head) = data[rc++];
		GALIL1800_CARD_WB_SET_PRESENT_FLAG(wbuff_p);
		galil1800_card_wbuffer_inc_pt(wbuff_p, wbuff_p->head);
		if(GALIL1800_CARD_WB_FULL(wbuff_p))
			break;
	}
	wake_up(&wbuff_p->read_wait); /* Don't remove this! */
	return rc;
}

/* Add a pause of ms milliseconds to buffer.
   Return 0 if success, -EINVAL if pause command is disable
   or -EAGAIN if there is no space into the buffer */
static int galil1800_card_wbuffer_add_pause(
		struct galil1800_card_wbuffer *wbuff_p, unsigned long ms)
{
	int rc = sizeof(unsigned long) - 1;
	unsigned long j = jiffies;
	unsigned char *head;

	if(!(wbuff_p->pause_enable)) {
		printk(KERN_WARNING PFX	"Cannot insert pause"
			" because pseudo pause command is disable\n");
		return -EINVAL;
	}

	GALIL_PRINT_DEBUG("Try to add pause = %lu, (0x%lx)\n", ms, ms);

	head = wbuff_p->head;
	do {
		if(GALIL1800_CARD_WB_TEST_PRESENT_FLAG(wbuff_p, head)) {
			printk(KERN_WARNING PFX "Add pause failed: "
					"no space in kernel buffer\n");
			return -EAGAIN;
		}
		galil1800_card_wbuffer_inc_pt(wbuff_p, head);
	}while(rc-- > 0);

	j+=((ms*HZ)/1000); /* Conv. in jiffies and sum with current jiffies */

	rc = sizeof(unsigned long) - 1;

	do {
		BUG_ON( GALIL1800_CARD_WB_FULL(wbuff_p) );
		*(wbuff_p->head) =
			(unsigned char)( (j >> (rc*8)) & 0x000000ff );
		GALIL_PRINT_DEBUG("Add pause wake up byte %d = %x\n", rc,
							*(wbuff_p->head));
		GALIL1800_CARD_WB_SET_PAUSE_FLAG(wbuff_p);
		GALIL1800_CARD_WB_SET_PRESENT_FLAG(wbuff_p);
		galil1800_card_wbuffer_inc_pt(wbuff_p, wbuff_p->head);
	}while(rc-- > 0);
	wake_up(&wbuff_p->read_wait);
	return 0;
}

static
inline void __galil1800_add_wbuff_event(struct galil1800_state *s,
					unsigned char wbuff_flag)
{
	unsigned char curr_size_p10; /* current size in per 10 value */
	unsigned char curr_size_p100; /* current size in per 100 value */

	curr_size_p100 =
		(GALIL1800_CARD_WB_CURRENT_SIZE(&s->card_wbuff)*100) /
							(s->card_wbuff.size);
	curr_size_p10 = curr_size_p100/10;

	if(s->card_wbuff.last_event_opt_flag==wbuff_flag &&
			s->card_wbuff.last_size_p10==curr_size_p10)
		return; /* Yet reported */

	s->card_wbuff.last_size_p10 = curr_size_p10;
	s->card_wbuff.last_event_opt_flag=wbuff_flag;

	switch(curr_size_p10) {
		case 9:
		case 10:
			galil1800_event_buff_add_wbuff_over_90p(
					&s->event_buff, curr_size_p100);
			break;
		case 0:
			if(galil1800_event_buff_add_wbuff_less_10p(
						&s->event_buff, curr_size_p100))
				break;
		case 1:
			if(galil1800_event_buff_add_wbuff_less_20p(
						&s->event_buff, curr_size_p100))
				break;
		case 2:
		case 3:
		case 4:
			galil1800_event_buff_add_wbuff_less_50p(
						&s->event_buff, curr_size_p100);
	};
}

#define galil1800_add_wbuff_event(_s) do {                          \
	unsigned char _wbuff_flag =                                 \
		galil1800_event_get_wbuff_flag(&(_s)->event_buff);  \
	if(_wbuff_flag) /* one or more event report enabled */      \
		__galil1800_add_wbuff_event(_s, _wbuff_flag);       \
}while(0)


/* Write from kernel write buffer to hardware fifo
   Used with pause-cmd disabled */
#define galil1800_card_wbuffer_write_to_fifo(_s, _wbuff_p) do {                \
	int _burst;                                                            \
	switch(__galil1800_may_write(_s)) {                                    \
		case 0: /* Sleep */                                            \
			while(__galil1800_may_write(_s)==0) {                  \
				GALIL_ONE_PRINT_DEBUG("card_wbuffer_flush "    \
					"going to sleep (fifo full)\n");       \
				wait_event_timeout((_wbuff_p)->read_wait,      \
						__galil1800_may_write(_s)>0,   \
						(_s)->write_delay );           \
			}                                                      \
			break;                                                 \
		case 1: /* One slot free */                                    \
			outb(*((_wbuff_p)->tail), GALIL1800_WRITE_REG(_s));    \
			GALIL1800_CARD_WB_CLEAR_PRESENT_FLAG(_wbuff_p);        \
			galil1800_card_wbuffer_inc_pt(_wbuff_p,                \
							(_wbuff_p)->tail);     \
			break;                                                 \
		case 2: /* Almost half fifo free */                            \
			_burst=0;                                              \
			do {                                                   \
				outb(*((_wbuff_p)->tail),                      \
					GALIL1800_WRITE_REG(_s));              \
				GALIL1800_CARD_WB_CLEAR_PRESENT_FLAG(_wbuff_p);\
				galil1800_card_wbuffer_inc_pt(_wbuff_p,        \
							(_wbuff_p)->tail);     \
				_burst++;                                      \
			}while(_burst < GALIL1800_HALF_FULL_WRITE_BURST &&     \
				GALIL1800_CARD_WB_NOT_EMPTY(_wbuff_p));        \
			NO_GALIL_PRINT_DEBUG("card_wbuffer_flush: burst=%d\n", \
								_burst);       \
			break;                                                 \
	};                                                                     \
}while(0)


/* Write from kernel write buffer to hardware fifo
   Used with pause-cmd enabled */
#define galil1800_card_wbuffer_write_to_fifo_with_pause_cmd(_s, _wbuff_p) do { \
	int _burst;                                                            \
	switch(__galil1800_may_write(_s)) {                                    \
		case 0: /* Sleep */                                            \
			while(__galil1800_may_write(_s)==0) {                  \
				GALIL_ONE_PRINT_DEBUG("card_wbuffer_flush "    \
					"going to sleep (fifo full)\n");       \
				wait_event_timeout((_wbuff_p)->read_wait,      \
						__galil1800_may_write(_s)>0,   \
						(_s)->write_delay );           \
			}                                                      \
			break;                                                 \
		case 1: /* One slot free */                                    \
			outb(*((_wbuff_p)->tail), GALIL1800_WRITE_REG(_s));    \
			GALIL1800_CARD_WB_CLEAR_PRESENT_FLAG(_wbuff_p);        \
			galil1800_card_wbuffer_inc_pt(_wbuff_p,                \
							(_wbuff_p)->tail);     \
			break;                                                 \
		case 2: /* Almost half fifo free */                            \
			_burst=0;                                              \
			do {                                                   \
	                        if(GALIL1800_CARD_WB_TEST_PAUSE(_wbuff_p))     \
					break;                                 \
				outb(*((_wbuff_p)->tail),                      \
					GALIL1800_WRITE_REG(_s));              \
				GALIL1800_CARD_WB_CLEAR_PRESENT_FLAG(_wbuff_p);\
				galil1800_card_wbuffer_inc_pt(_wbuff_p,        \
							(_wbuff_p)->tail);     \
				_burst++;                                      \
			}while(_burst < GALIL1800_HALF_FULL_WRITE_BURST &&     \
				GALIL1800_CARD_WB_NOT_EMPTY(_wbuff_p));        \
			NO_GALIL_PRINT_DEBUG("card_wbuffer_flush: burst=%d\n", \
								_burst);       \
			break;                                                 \
	};                                                                     \
}while(0)


/* Function called by a workqueue: write data from kernel buffer to hardware fifo.
   This is used if "pause" command is disabled */
void galil1800_card_wbuffer_flush(void *data)
{
	struct galil1800_card_wbuffer *wbuff_p =
		(struct galil1800_card_wbuffer *)data;
	struct galil1800_state *s =
		container_of(wbuff_p, struct galil1800_state, card_wbuff);

	while(1) {
		/* Wait to data in buffer */
		while(!GALIL1800_CARD_WB_NOT_EMPTY(wbuff_p)) {
			if(!wbuff_p->run)
				return;
			wake_up_interruptible(&wbuff_p->write_wait);
			NO_GALIL_ONE_PRINT_DEBUG("card_wbuffer_flush "
					"going to sleep (buffer empty)\n");
			wait_event(wbuff_p->read_wait,
				GALIL1800_CARD_WB_NOT_EMPTY(wbuff_p) ||
							 !wbuff_p->run);
		}
		galil1800_add_wbuff_event(s);

		/* Now there is data in buffer, so: */

		do { /* While buffer is not empty... */
			galil1800_card_wbuffer_write_to_fifo(s, wbuff_p);
			galil1800_add_wbuff_event(s);
		}while(GALIL1800_CARD_WB_NOT_EMPTY(wbuff_p));
	wake_up_interruptible(&wbuff_p->write_wait);
	}
}


#define galil1800_card_wbuff_prepare_pause(_wbuff_p) do {                 \
	int _pause_bytes = 0;                                             \
	do {                                                              \
		(_wbuff_p)->pause &= 0xffffff00;                          \
		(_wbuff_p)->pause |= *((_wbuff_p)->tail);                 \
		if(_pause_bytes++ < 3)                                    \
			(_wbuff_p)->pause = (_wbuff_p)->pause << 8;       \
		GALIL1800_CARD_WB_CLEAR_PAUSE_FLAG(_wbuff_p);             \
		GALIL1800_CARD_WB_CLEAR_PRESENT_FLAG(_wbuff_p);           \
		galil1800_card_wbuffer_inc_pt(_wbuff_p, (_wbuff_p)->tail);\
		while(!GALIL1800_CARD_WB_NOT_EMPTY(_wbuff_p)) {           \
			GALIL_ONE_PRINT_DEBUG("card_wbuffer_flush going"  \
				" to sleep (wait for pause byte)\n");     \
			wait_event((_wbuff_p)->read_wait,                 \
				GALIL1800_CARD_WB_NOT_EMPTY(_wbuff_p));   \
		}                                                         \
	}while(_pause_bytes < 4);                                         \
	GALIL_PRINT_DEBUG("Read pause %lu ticks = %lx\n",                 \
			(_wbuff_p)->pause, (_wbuff_p)->pause);            \
}while(0)


/* Function called by a workqueue:
   write data from kernel buffer to hardware fifo.
   This is used if "pause" command is enabled */
void galil1800_card_wbuffer_flush_whit_pause_cmd(void *data)
{
	struct galil1800_card_wbuffer *wbuff_p =
		(struct galil1800_card_wbuffer *)data;
	struct galil1800_state *s =
		container_of(wbuff_p, struct galil1800_state, card_wbuff);

	while(1) {
		/* Wait to data in buffer */
		while(!GALIL1800_CARD_WB_NOT_EMPTY(wbuff_p)) {
			if(!wbuff_p->run)
				return;
			wake_up_interruptible(&wbuff_p->write_wait);
			NO_GALIL_ONE_PRINT_DEBUG("card_wbuffer_flush "
					"going to sleep (buffer empty)\n");
			wait_event(wbuff_p->read_wait,
				GALIL1800_CARD_WB_NOT_EMPTY(wbuff_p) ||
							!wbuff_p->run);
		}
		galil1800_add_wbuff_event(s);

		/* Now there is data in buffer, so: */
		do { /* While buffer is not empty... */
			if(GALIL1800_CARD_WB_TEST_PAUSE(wbuff_p)) {
				galil1800_card_wbuff_prepare_pause(wbuff_p);
				galil1800_add_pause_timer(wbuff_p);

				/* Wait to pause timer end */
				while(wbuff_p->pause > 0) {
					GALIL_PRINT_DEBUG("card_wbuffer_flush"
						" going to sleep (pause)\n");
					wait_event(wbuff_p->read_wait,
							!wbuff_p->pause);
				}
				break;
			}

			galil1800_card_wbuffer_write_to_fifo_with_pause_cmd(
								s, wbuff_p);
			galil1800_add_wbuff_event(s);
		}while(GALIL1800_CARD_WB_NOT_EMPTY(wbuff_p));
		wake_up_interruptible(&wbuff_p->write_wait);
	}
}

/* Write call without kernel buffer */
static ssize_t galil1800_card_unbuffered_write(struct file * filp,
		const char __user * buf, size_t size, loff_t *poff)
{
	struct galil1800_state *dev =
		(struct galil1800_state *)filp->private_data;
	char *_buf;
	ssize_t request;
	int rc=0;

	GALIL_PRINT_DEBUG_PID("\"card unbuffered write\" called\n");

	if (!size)
		return 0;
	if (!buf || size < 0)
		return -EINVAL;

	request = (size < GALIL1800_WRITE_FIFO_SIZE ?
			size : GALIL1800_WRITE_FIFO_SIZE);

	_buf = kmalloc(request, GFP_KERNEL);
	if (!_buf) {
		printk(KERN_ERR PFX "card write: alloc _buf failed\n");
		return -ENOMEM;
	}

	if (copy_from_user(_buf, buf, request)) {
		rc = -EFAULT;
		goto out;
	}

	rc = galil1800_send(dev, _buf, request);

out:
	kfree(_buf);
	return rc;
}

/* Write call with kernel buffer */
static ssize_t galil1800_card_buffered_write(struct file * filp,
		const char __user * buf, size_t size, loff_t *poff)
{
	struct galil1800_state *dev =
		(struct galil1800_state *)filp->private_data;
	char *_buf;
	ssize_t request;
	int rc=0;

	GALIL_PRINT_DEBUG_PID("\"card buffered write\" called\n");

	if (!size)
		return 0;
	if (!buf || size < 0)
		return -EINVAL;

	request = (size < dev->card_wbuff.size ? size : dev->card_wbuff.size);

	_buf = kmalloc(request, GFP_KERNEL);
	if (!_buf) {
		printk(KERN_ERR PFX "card write: alloc _buf failed\n");
		return -ENOMEM;
	}

	if (copy_from_user(_buf, buf, request)) {
		rc = -EFAULT;
		goto out;
	}

	rc = galil1800_card_wbuffer_add(&dev->card_wbuff, _buf, request,
						filp->f_flags & O_NONBLOCK);

out:
	kfree(_buf);
	return rc;
}


static ssize_t galil1800_card_read(struct file * filp,
		char __user * buf,size_t size, loff_t *poff)
{
	struct galil1800_state *s =
		(struct galil1800_state *) filp->private_data;
	char *_buf;
	ssize_t count;
	ssize_t request = (size < GALIL1800_READ_FIFO_SIZE ?
					size : GALIL1800_READ_FIFO_SIZE);
	int rc;

	GALIL_PRINT_DEBUG_PID("\"card read\" called\n");

	/* Wait to data arrive */
	while(__galil1800_may_read(s)==0) {
		if(filp->f_flags & O_NONBLOCK)
			return -EAGAIN;
		NO_GALIL_PRINT_DEBUG("card read going to sleep"
						" (read fifo empty)\n");
		rc = wait_event_interruptible_timeout(s->card_read_wait,
				__galil1800_may_read(s)>0, s->read_delay);
		if(rc == -ERESTARTSYS) {
			GALIL_PRINT_DEBUG("card read interrupted\n");
			return rc;
		}
	}
	NO_GALIL_PRINT_DEBUG("galil1800_card_read wake up!\n");

	_buf = kmalloc(request, GFP_KERNEL);
	if (!_buf) {
		printk(KERN_ERR PFX "card read: alloc _buf failed\n");
		return -ENOMEM;
	}

	count = galil1800_recv(s, _buf, request);

	rc = 0;
	if (count)
		rc = copy_to_user(buf, _buf, count);

	kfree(_buf);
	return (rc > 0 ? -EFAULT : count);
}

static
int galil1800_card_fasync(int fd, struct file *filp, int mode)
{
/*
    struct galil1800_state *dev =
            (struct galil1800_state *) filp->private_data;
    return fasync_helper(fd, filp, mode, &dev->async_queue);
*/
    /* TODO Aggiungere "struct faync_struct *async_queue"
            nella struct galil1800_state (inizializato a NULL??),
            poi preoccuparsi di trovare qualcosa che segnali quando
            la fifo di lettura ha dei dati pronti e quindi usi la funzione:
            if(dev->async_queue) kill_fasync(&dev->async_queue, SIGIO, POLL_IN);
            Infine nella chiamata di release aggiungere:
                galil1800_card_fasync(-1,filp,0);
            [ vedi Rubini p. 169-171 ]*/
    return -EFAULT;
}

static
int galil1800_card_open(struct inode * inode, struct file * filp)
{
	int ret;
	struct galil1800_state *s;
	struct galil1800_devfile_state *df_s;

	s = container_of(inode->i_cdev, struct galil1800_state,
			devfile_state[GALIL1800_CARD_DEVICE].cdev);
	filp->private_data = s;
	df_s = &(s->devfile_state[GALIL1800_CARD_DEVICE]);

	/* wait for device to become free */
	GALIL_PRINT_DEBUG_PID("\"card open\" started\n");

	if(down_interruptible(&df_s->open_sem)) {
		GALIL_PRINT_DEBUG_PID("\"card open\": \"down\" interrupted\n");
		return -ERESTARTSYS;
	}

	/* Whit this type of check,
	   one process can open for read and another for write */

	while (df_s->open_mode & filp->f_mode) {
		if (filp->f_flags & O_NONBLOCK) { /* O_NONBLOCK == O_NDELAY */
			up(&df_s->open_sem);
			GALIL_PRINT_DEBUG_PID("\"card open\" return:"
				" O_NONBLOCK is set and device is busy\n");
			return -EAGAIN;
		}
		up(&df_s->open_sem);
		GALIL_ONE_PRINT_DEBUG_PID("\"card open\" going to sleep\n");
		if(wait_event_interruptible(df_s->open_wait,
					!(df_s->open_mode & filp->f_mode))) {
			GALIL_PRINT_DEBUG_PID("\"card open\":"
					" \"wait_event\" interrupted\n");
			return -ERESTARTSYS;
		}
		if(down_interruptible(&df_s->open_sem)) {
			GALIL_PRINT_DEBUG_PID("\"card open\":"
					" \"down\" interrupted\n");
			return -ERESTARTSYS;
		}
	}

	/* Only one can open to write... */
	if(filp->f_mode & FMODE_WRITE) {
		ret=init_galil1800_card_wbuffer_open_dev(&s->card_wbuff);
		if(ret < 0) {
			up(&df_s->open_sem);
			return ret;
		}
	}

	ret = galil1800_init_controller_open_step(s);
	if(ret < 0) {
		if(filp->f_mode & FMODE_WRITE)
			clean_galil1800_card_wbuffer_close_dev(&s->card_wbuff);
		up(&df_s->open_sem);
		return ret;
	}

	df_s->open_mode |= filp->f_mode & (FMODE_READ | FMODE_WRITE);

	up(&df_s->open_sem);

	nonseekable_open(inode,filp);

	GALIL_PRINT_DEBUG_PID("\"card open\" ended\n");

	return 0;
}


static int galil1800_card_release(struct inode * inode, struct file * filp)
{
	struct galil1800_state *s =
		(struct galil1800_state *)filp->private_data;
	struct galil1800_devfile_state *df_s =
		&(s->devfile_state[GALIL1800_CARD_DEVICE]);

	GALIL_PRINT_DEBUG_PID("\"card release\" started\n");
	if(down_interruptible(&df_s->open_sem)) {
		GALIL_PRINT_DEBUG_PID("\"card release\":"
				" \"down\" interrupted\n");
		return -ERESTARTSYS;
	}

	/* Only one can open to write... */
	if(filp->f_mode & FMODE_WRITE) {
		GALIL_PRINT_DEBUG_PID("\"card release\":"
				" start to clean wbuff\n");
		clean_galil1800_card_wbuffer_close_dev(&s->card_wbuff);
		GALIL_PRINT_DEBUG_PID("\"card release\":"
				" end to clean wbuff\n");
	}

	galil1800_clean_controller_open_step(s);

	df_s->open_mode &= (~filp->f_mode) & (FMODE_READ|FMODE_WRITE);

	up(&df_s->open_sem);
	wake_up_interruptible_sync(&df_s->open_wait);
	GALIL_PRINT_DEBUG_PID("\"card release\" ended:"
			" wake up of sleeping open\n");
	return 0;
}

static
long galil1800_card_unlocked_ioctl(struct file *filp,
			unsigned int cmd, unsigned long arg)
{
	struct galil1800_state *dev =
		(struct galil1800_state *) filp->private_data;

	/* Check command */
	if(_IOC_TYPE(cmd) != GALIL1800_IOCTL_MAGIC)
		return -ENOTTY;
	if(_IOC_NR(cmd) < GALIL1800_IOCTL_CARD_MINNR ||
			_IOC_NR(cmd) > GALIL1800_IOCTL_CARD_MAXNR)
		return -ENOTTY;

	/* Check argument:
	   After this check, we safely use:
	   __copy_to_user or __copy_from_user
	   __put_user or __get_user          */
	if(_IOC_DIR(cmd) & _IOC_READ) {
		/* Read request, so check if I can write to user space */
		NO_GALIL_PRINT_DEBUG("ioctl card: read request (cmd: %d)\n",
								_IOC_NR(cmd));
		if(!(access_ok(VERIFY_WRITE,(void __user *)arg,_IOC_SIZE(cmd))))
			return -EFAULT;
	}
	else if(_IOC_DIR(cmd) & _IOC_WRITE) {
		/* Write request, so check if I can read from user space */
		NO_GALIL_PRINT_DEBUG("ioctl card: write request (cmd: %d)\n",
								_IOC_NR(cmd));
		if(!(access_ok(VERIFY_READ,(void __user *)arg,_IOC_SIZE(cmd))))
			return -EFAULT;
	}

    switch (cmd) {
		case GALIL1800_IOCTL_RESET_HARDWARE:
		{
			/* To reset the controller, clear the FIFO,
			   then send the RS command;
			If the controller is not responding,
			you can use this hardware reset */
			galil1800_reset_fifo_rw(dev);
			galil1800_reset_hardware(dev);
		}
		break;
		case GALIL1800_IOCTL_RESET_FIFO_R:
		{
			galil1800_reset_fifo_r(dev);
		}
		break;
		case GALIL1800_IOCTL_RESET_FIFO_W:
		{
			galil1800_reset_fifo_w(dev);
		}
		break;
		case GALIL1800_IOCTL_RESET_FIFO_RW:
		{
			galil1800_reset_fifo_rw(dev);
		}
		break;
		case GALIL1800_IOCTL_UNSAFE_WRITE_STOP:
		{
			galil1800_unsafe_write_stop(&dev->card_wbuff);
		}
		break;
		case GALIL1800_IOCTL_GET_SERIAL_NUMBER:
		{
			__copy_to_user((char*)arg,dev->serial,_IOC_SIZE(cmd));
		}
		break;
		case GALIL1800_IOCTL_SET_PAUSE:
		{
			/* In this way, I'm sure that this call isn't in
			   parallel with write:
			   only one process can open for write! */
			if(filp->f_mode & FMODE_WRITE)
				return galil1800_card_wbuffer_add_pause(
							&dev->card_wbuff, arg);
			else {
				printk(KERN_NOTICE PFX "ioctl on card device:"
					" to set pause needs write access!\n");
				return -EACCES;
			}
		}
		break;
		case GALIL1800_IOCTL_SET_WRITE_DELAY:
		{
			unsigned int temp = (unsigned int)arg;
			/* Now check if value set for write_delay from user
			is correct and eventually correct it
			(see in galil1800_io.h)
			After convert it from ms to ticks!!
			Return 0 if ok, -EINVAL if reset to default */
			int ret=galil1800_write_delay_check_and_convert(&temp);
			dev->write_delay = temp;
			return ret;
		}
		break;
		case GALIL1800_IOCTL_SET_READ_DELAY:
		{
			unsigned int temp = (unsigned int)arg;
			/* Now check if value set for read_delay from user
			is correct and eventually correct it
			(see in galil1800_io.h)
			After convert it from ms to ticks!!
			Return 0 if ok, -EINVAL if reset to default */
			int ret=galil1800_read_delay_check_and_convert(&temp);
			dev->read_delay = temp;
			return ret;
		}
		break;
		case GALIL1800_IOCTL_SET_WRITE_TIMEOUT:
		{
			unsigned int temp = (unsigned int)arg;
			/* Now check if value set for write_timeout from user
			is correct and eventually correct it
			(see in galil1800_io.h)
			After convert it from ms to ticks!!
			Return 0 if ok, -EINVAL if reset to default */
			int ret=galil1800_write_timeout_check_and_convert(&temp);
			dev->write_timeout = temp;
			return ret;
		}
		break;
		case GALIL1800_IOCTL_SET_READ_TIMEOUT:
		{
			unsigned int temp = (unsigned int)arg;
			/* Now check if value set for read_timeout from user
			is correct and eventually correct it
			(see in galil1800_io.h)
			After convert it from ms to ticks!!
			Return 0 if ok, -EINVAL if reset to default */
			int ret = galil1800_read_timeout_check_and_convert(&temp);
			dev->read_timeout = temp;
			return ret;
		}
		break;
		default:
			return -ENOTTY; /* cmd not valid */
	}
	return 0; /* Success */
}


struct file_operations galil1800_card_unbuffered_fops = {
	.owner = THIS_MODULE,
	.read = galil1800_card_read,
	.write = galil1800_card_unbuffered_write,
	.fasync = galil1800_card_fasync,
	.unlocked_ioctl = galil1800_card_unlocked_ioctl,
	.open = galil1800_card_open,
	.release = galil1800_card_release,
	.llseek = no_llseek
};

struct file_operations galil1800_card_buffered_fops = {
	.owner = THIS_MODULE,
	.read = galil1800_card_read,
	.write = galil1800_card_buffered_write,
	.fasync = galil1800_card_fasync,
	.unlocked_ioctl = galil1800_card_unlocked_ioctl,
	.open = galil1800_card_open,
	.release = galil1800_card_release,
	.llseek = no_llseek
};

static
ssize_t galil1800_sysfs_card_wbuff_sz_show(struct class_device *cd, char *buf)
{
	struct galil1800_state *s = container_of(
		(struct galil1800_devfile_state *)(cd->class_data),
		struct galil1800_state, devfile_state[GALIL1800_CARD_DEVICE]);
	return sprintf(buf, "%u\n", s->card_wbuff.size);
}

static
ssize_t galil1800_sysfs_card_wbuff_sz_store(struct class_device *cd,
						const char *buf, size_t count)
{
	unsigned int last_size;

	struct galil1800_devfile_state *df_s =
		(struct galil1800_devfile_state *)(cd->class_data);
	struct galil1800_state *s = container_of(df_s,
		struct galil1800_state, devfile_state[GALIL1800_CARD_DEVICE]);

	GALIL_PRINT_DEBUG("galil1800_sysfs_card_wbuff_sz_store, buf: \"%s\"\n",
									buf);

	if(down_interruptible(&df_s->open_sem)) {
		GALIL_PRINT_DEBUG_PID("\"card_wbuff_sz_store\": \"down\""
							" interrupted\n");
		return -ERESTARTSYS;
	}

	if(df_s->open_mode) {
		up(&df_s->open_sem);
		printk(KERN_ERR PFX "card_wbuff_sz not modified."
					" There are files opened\n");
		return -EBUSY;
	}
	if(count>10 || strlen(buf)!=count) {
		up(&df_s->open_sem);
		printk(KERN_ERR PFX "card_wbuff_sz not modified."
						" Too much characters\n");
		return -EINVAL;
	}

	last_size = s->card_wbuff.size;

	sscanf(buf, "%u", &s->card_wbuff.size);

	galil1800_card_wbuff_size_check(&s->card_wbuff.size);

	if(s->card_wbuff.size==0)
		df_s->cdev.ops = &galil1800_card_unbuffered_fops;
	else
		df_s->cdev.ops = &galil1800_card_buffered_fops;

	up(&df_s->open_sem);

	return count;
}

static
ssize_t galil1800_sysfs_pause_enable_show(struct class_device *cd, char *buf)
{
	struct galil1800_state *s = container_of(
		(struct galil1800_devfile_state *)(cd->class_data),
		struct galil1800_state, devfile_state[GALIL1800_CARD_DEVICE]);
	return sprintf(buf, "%u\n", s->card_wbuff.pause_enable);
}

static
ssize_t galil1800_sysfs_pause_enable_store(struct class_device *cd,
					const char *buf, size_t count)
{
	struct galil1800_devfile_state *df_s =
		(struct galil1800_devfile_state *)(cd->class_data);
	struct galil1800_state *s = container_of(df_s,
		struct galil1800_state, devfile_state[GALIL1800_CARD_DEVICE]);

	GALIL_PRINT_DEBUG("galil1800_sysfs_pause_enable_store, buf: \"%s\"\n",
									buf);

	if(down_interruptible(&df_s->open_sem)) {
		GALIL_PRINT_DEBUG_PID("\"pause_enable_store\":"
					" \"down\" interrupted\n");
		return -ERESTARTSYS;
	}

	if(df_s->open_mode) {
		up(&df_s->open_sem);
		printk(KERN_ERR PFX "pause_enable not modified."
					" There are files opened\n");
		return -EBUSY;
	}
	if(count>10 || strlen(buf)!=count) {
		up(&df_s->open_sem);
		printk(KERN_ERR PFX "pause_enable not modified."
					" Too much characters\n");
		return -EINVAL;
	}

	sscanf(buf, "%u", &s->card_wbuff.pause_enable);

	galil1800_card_parameter_check(&s->card_wbuff.pause_enable,
						&s->card_wbuff.size);

	if(s->card_wbuff.pause_enable)
		PREPARE_WORK(&s->card_wbuff.work,
		galil1800_card_wbuffer_flush_whit_pause_cmd, &s->card_wbuff);
	else
		PREPARE_WORK(&s->card_wbuff.work, galil1800_card_wbuffer_flush,
								&s->card_wbuff);

	up(&df_s->open_sem);

	return count;
}

static
ssize_t galil1800_sysfs_write_delay_show(struct class_device *cd, char *buf)
{
	struct galil1800_state *s = container_of(
		(struct galil1800_devfile_state *)(cd->class_data),
		struct galil1800_state, devfile_state[GALIL1800_CARD_DEVICE]);
	return sprintf(buf, "%u\n", (s->write_delay*1000)/HZ); /* in ms */
}

static
ssize_t galil1800_sysfs_write_delay_store(struct class_device *cd,
						const char *buf, size_t count)
{
	unsigned int temp;
	struct galil1800_state *s = container_of(
		(struct galil1800_devfile_state *)(cd->class_data),
		struct galil1800_state, devfile_state[GALIL1800_CARD_DEVICE]);

	GALIL_PRINT_DEBUG("galil1800_sysfs_write_delay_store, buf: \"%s\"\n",
									 buf);

	if(count > 5 || strlen(buf)!=count) {
		printk(KERN_ERR PFX "write_delay not modified."
					" Too much characters\n");
		return -EINVAL;
	}

	sscanf(buf, "%u", &temp);

	galil1800_write_delay_check_and_convert(&temp);

	s->write_delay = temp;

	return count;
}

static
ssize_t galil1800_sysfs_read_delay_show(struct class_device *cd, char *buf)
{
	struct galil1800_state *s = container_of(
		(struct galil1800_devfile_state *)(cd->class_data),
		struct galil1800_state, devfile_state[GALIL1800_CARD_DEVICE]);
	return sprintf(buf, "%u\n", (s->read_delay*1000)/HZ); /* in ms */
}

static
ssize_t galil1800_sysfs_read_delay_store(struct class_device *cd,
						const char *buf, size_t count)
{
	unsigned int temp;
	struct galil1800_state *s = container_of(
		(struct galil1800_devfile_state *)(cd->class_data),
		struct galil1800_state, devfile_state[GALIL1800_CARD_DEVICE]);

	GALIL_PRINT_DEBUG("galil1800_sysfs_read_delay_store, buf: \"%s\"\n",
									 buf);

	if(count > 5 || strlen(buf)!=count) {
		printk(KERN_ERR PFX "read_delay not modified."
					" Too much characters\n");
		return -EINVAL;
	}

	sscanf(buf, "%u", &temp);

	galil1800_read_delay_check_and_convert(&temp);

	s->read_delay = temp;

	return count;
}

static
ssize_t galil1800_sysfs_write_timeout_show(struct class_device *cd, char *buf)
{
	struct galil1800_state *s = container_of(
		(struct galil1800_devfile_state *)(cd->class_data),
		struct galil1800_state, devfile_state[GALIL1800_CARD_DEVICE]);
	return sprintf(buf, "%u\n", (s->write_timeout*1000)/HZ); /* in ms */
}

static
ssize_t galil1800_sysfs_write_timeout_store(struct class_device *cd,
						const char *buf, size_t count)
{
	unsigned int temp;
	struct galil1800_state *s = container_of(
		(struct galil1800_devfile_state *)(cd->class_data),
		struct galil1800_state, devfile_state[GALIL1800_CARD_DEVICE]);

	GALIL_PRINT_DEBUG("galil1800_sysfs_write_timeout_store, buf: \"%s\"\n",
									 buf);

	if(count > 5 || strlen(buf)!=count) {
		printk(KERN_ERR PFX "write_timeout not modified."
					" Too much characters\n");
		return -EINVAL;
	}

	sscanf(buf, "%u", &temp);

	galil1800_write_timeout_check_and_convert(&temp);

	s->write_timeout = temp;

	return count;
}

static
ssize_t galil1800_sysfs_read_timeout_show(struct class_device *cd, char *buf)
{
	struct galil1800_state *s = container_of(
		(struct galil1800_devfile_state *)(cd->class_data),
		struct galil1800_state, devfile_state[GALIL1800_CARD_DEVICE]);
	return sprintf(buf, "%u\n", (s->read_timeout*1000)/HZ); /* in ms */
}

static ssize_t galil1800_sysfs_read_timeout_store(struct class_device *cd,
						const char *buf, size_t count)
{
	unsigned int temp;
	struct galil1800_state *s = container_of(
		(struct galil1800_devfile_state *)(cd->class_data),
		struct galil1800_state, devfile_state[GALIL1800_CARD_DEVICE]);

	GALIL_PRINT_DEBUG("galil1800_sysfs_read_timeout_store, buf: \"%s\"\n",
									 buf);

	if(count > 5 || strlen(buf)!=count) {
		printk(KERN_ERR PFX "read_timeout not modified."
					" Too much characters\n");
		return -EINVAL;
	}

	sscanf(buf, "%u", &temp);

	galil1800_read_timeout_check_and_convert(&temp);

	s->read_timeout = temp;

	return count;
}

/* Add card cdev device and register it in sysfs */
int galil1800_card_cdev_add(struct cdev *cdev_p,
				unsigned int major, unsigned int minor)
{
	int ret;
	const char *err;
	struct galil1800_state *s = container_of(cdev_p,
		struct galil1800_state,
		devfile_state[GALIL1800_CARD_DEVICE].cdev);
	struct class_device *cd =
		 &s->devfile_state[GALIL1800_CARD_DEVICE].class_dev;

	if(s->card_wbuff.size==0)
		cdev_init(cdev_p, &galil1800_card_unbuffered_fops);
	else
		cdev_init(cdev_p, &galil1800_card_buffered_fops);

	cdev_p->owner = THIS_MODULE;

	ret=cdev_add(cdev_p, MKDEV(major, minor), 1);
	if (ret) {
		printk(KERN_INFO PFX "cannot add card cdev dev=(%d,%d)\n",
								major, minor);
		return ret;
	}

	ret=galil1800_sysfs_class_dev_register(s, GALIL1800_CARD_DEVICE);
	if(ret) {
		printk(KERN_INFO PFX "sysfs class device register"
					" failed for card device\n");
		goto err_class_dev_reg;
	}


	s->card_wbuff.cd_attr_size.attr.name = "card_wbuff_sz";
	s->card_wbuff.cd_attr_size.attr.owner = THIS_MODULE;
	s->card_wbuff.cd_attr_size.attr.mode = S_IWUSR | S_IRUGO;
	s->card_wbuff.cd_attr_size.show  = galil1800_sysfs_card_wbuff_sz_show;
	s->card_wbuff.cd_attr_size.store = galil1800_sysfs_card_wbuff_sz_store;
	ret=class_device_create_file(cd, &s->card_wbuff.cd_attr_size);
	if(ret) {
		err = s->card_wbuff.cd_attr_size.attr.name;
		goto err_attr_size;
	}

	s->card_wbuff.cd_attr_pause_enable.attr.name = "pause_enable";
	s->card_wbuff.cd_attr_pause_enable.attr.owner = THIS_MODULE;
	s->card_wbuff.cd_attr_pause_enable.attr.mode = S_IWUSR | S_IRUGO;
	s->card_wbuff.cd_attr_pause_enable.show  =
				galil1800_sysfs_pause_enable_show;
	s->card_wbuff.cd_attr_pause_enable.store =
				galil1800_sysfs_pause_enable_store;
	ret=class_device_create_file(cd, &s->card_wbuff.cd_attr_pause_enable);
	if(ret) {
		err = s->card_wbuff.cd_attr_pause_enable.attr.name;
		goto err_attr_pause_enable;
	}

	s->cd_attr_write_delay.attr.name = "write_delay";
	s->cd_attr_write_delay.attr.owner = THIS_MODULE;
	s->cd_attr_write_delay.attr.mode = S_IWUSR | S_IRUGO;
	s->cd_attr_write_delay.show  = galil1800_sysfs_write_delay_show;
	s->cd_attr_write_delay.store = galil1800_sysfs_write_delay_store;
	ret=class_device_create_file(cd,
			&s->cd_attr_write_delay);
	if(ret) {
		err = s->cd_attr_write_delay.attr.name;
		goto err_attr_write_delay;
	}

	s->cd_attr_read_delay.attr.name = "read_delay";
	s->cd_attr_read_delay.attr.owner = THIS_MODULE;
	s->cd_attr_read_delay.attr.mode = S_IWUSR | S_IRUGO;
	s->cd_attr_read_delay.show  = galil1800_sysfs_read_delay_show;
	s->cd_attr_read_delay.store = galil1800_sysfs_read_delay_store;
	ret=class_device_create_file(cd,
			&s->cd_attr_read_delay);
	if(ret) {
		err = s->cd_attr_read_delay.attr.name;
		goto err_attr_read_delay;
	}

	s->cd_attr_write_timeout.attr.name = "write_timeout";
	s->cd_attr_write_timeout.attr.owner = THIS_MODULE;
	s->cd_attr_write_timeout.attr.mode = S_IWUSR | S_IRUGO;
	s->cd_attr_write_timeout.show  = galil1800_sysfs_write_timeout_show;
	s->cd_attr_write_timeout.store = galil1800_sysfs_write_timeout_store;
	ret=class_device_create_file(cd,
			&s->cd_attr_write_timeout);
	if(ret) {
		err = s->cd_attr_write_timeout.attr.name;
		goto err_attr_write_timeout;
	}

	s->cd_attr_read_timeout.attr.name = "read_timeout";
	s->cd_attr_read_timeout.attr.owner = THIS_MODULE;
	s->cd_attr_read_timeout.attr.mode = S_IWUSR | S_IRUGO;
	s->cd_attr_read_timeout.show  = galil1800_sysfs_read_timeout_show;
	s->cd_attr_read_timeout.store = galil1800_sysfs_read_timeout_store;
	ret=class_device_create_file(cd,
			&s->cd_attr_read_timeout);
	if(ret) {
		err = s->cd_attr_read_timeout.attr.name;
		goto err_attr_read_timeout;
	}

	return 0;

err_attr_read_timeout:
	class_device_remove_file(cd, &s->cd_attr_write_timeout);
err_attr_write_timeout:
	class_device_remove_file(cd, &s->cd_attr_read_delay);
err_attr_read_delay:
	class_device_remove_file(cd, &s->cd_attr_write_delay);
err_attr_write_delay:
	class_device_remove_file(cd, &s->card_wbuff.cd_attr_pause_enable);
err_attr_pause_enable:
	class_device_remove_file(cd, &s->card_wbuff.cd_attr_size);
err_attr_size:
	printk(KERN_ERR PFX "cannot create file for class "
			"device attr \"%s\" in sysfs\n", err);
	galil1800_sysfs_class_dev_unregister(s, GALIL1800_CARD_DEVICE);
err_class_dev_reg:
	cdev_del(cdev_p);
	return ret;
}

/* Delete card cdev device and deregister it in sysfs */
void galil1800_card_cdev_del(struct cdev *cdev_p)
{
	struct galil1800_state *s = container_of(cdev_p,
		struct galil1800_state,
		devfile_state[GALIL1800_CARD_DEVICE].cdev);
	struct class_device *cd =
		 &s->devfile_state[GALIL1800_CARD_DEVICE].class_dev;

	galil1800_sysfs_class_dev_unregister(s, GALIL1800_CARD_DEVICE);

	class_device_remove_file(cd, &s->card_wbuff.cd_attr_size);
	class_device_remove_file(cd, &s->card_wbuff.cd_attr_pause_enable);
	class_device_remove_file(cd, &s->cd_attr_write_delay);
	class_device_remove_file(cd, &s->cd_attr_read_delay);
	class_device_remove_file(cd, &s->cd_attr_write_timeout);
	class_device_remove_file(cd, &s->cd_attr_read_timeout);

	cdev_del(cdev_p);

}
