/* Copyright (C) 2005-2006  Emiliano Betti and Marco Cesati
   Copyright (C) 2003-2004  Marco Cesati
   Copyright (C) 2002-2003  Shell Technologies s.r.l.

   Authors:
            Emiliano Betti <emilianobetti@gmail.com>
            Marco Cesati   <marcocesati@gmail.com>

   Driver for the Galil 1800 PCI motion controllers (http://www.galilmc.com/)

   This is a driver for the Linux kernel 2.6. It is based on an older driver
   for the kernel 2.4, which in turn was loosely inspired by Galil's own
   Linux driver, as well as by a bunch of PCI drivers in the kernel source code

   This program is free software; you can redistribute  it and/or modify it
   under  the terms of  the GNU General  Public License as published by the
   Free Software Foundation.
*/


#include "galil1800_io.h"
#include "galil1800_init.h"
#include "galil1800_sysfs.h"

#include <asm/uaccess.h>
#include <linux/interrupt.h>
#include <linux/poll.h>


/* This is the top half interrupt handler */
irqreturn_t galil1800_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	struct galil1800_state *dev = dev_id;
	unsigned long flags;
	unsigned char status, t;

	t = inb(GALIL1800_CONTROL_REG(dev));
	if(!(t & GALIL1800_IRQ_ENABLE_FLAG) || !(t & GALIL1800_IRQ_STATUS_FLAG))
		return IRQ_NONE; /* shared interrupt raised by someone else */

	status = inb(GALIL1800_IRQ_REG(dev));

	/* Here add in stack don't need to acquire the spinlock... */
	galil1800_irq_stack_add(&dev->irq_stack, status);

	spin_lock_irqsave(&dev->hwlock, flags);
	t = inb(GALIL1800_CONTROL_REG(dev)); /* need to re-read!! */
	outb(t | GALIL1800_IRQ_STATUS_FLAG, GALIL1800_CONTROL_REG(dev));
	spin_unlock_irqrestore(&dev->hwlock, flags);

	/* Schedule request for bottom half handling */
	tasklet_schedule(&dev->irq_tasklet);

	return IRQ_HANDLED;
}

/* This is the bottom half interrupt handler */
void galil1800_bottom_half_interrupt(unsigned long data) {
	struct galil1800_state *s = (struct galil1800_state *)data;
	struct galil1800_event_buffer *buff_p = &s->event_buff;

	/* Now add irq event from irq_stack to event buffer */
	if (galil1800_event_is_active(buff_p, GALIL1800_EVENT_IRQ)) {
		spin_lock_bh(&buff_p->lock_pt);
		/* I must retest because event device release */
		/* function can free the buffer! */
		if (galil1800_event_is_active(buff_p, GALIL1800_EVENT_IRQ)) {
			for_each_galil1800_irq_stack(&s->irq_stack) {
				__galil1800_event_buff_add(buff_p,
					*(s->irq_stack.tail),
					GALIL1800_EVENT_IRQ);
			}
		} else
			s->irq_stack.tail = s->irq_stack.head;
		spin_unlock_bh(&buff_p->lock_pt);
		wake_up_interruptible_sync(&buff_p->read_wait);
	} else
		s->irq_stack.tail = s->irq_stack.head;

}


/* Check if value set for event buffer size (*ev_buff_sz)
    is correct and eventually correct it
    Return 0 if no correct is needed, otherwise a positive value */
int galil1800_event_buff_size_check(unsigned int *ev_buff_sz)
{
	if(*ev_buff_sz < GALIL1800_EVENT_BUFF_SZ_MIN) {
		/* A circular buffer cannot work!! */
		*ev_buff_sz = GALIL1800_EVENT_BUFF_DEFAULT_SZ;
		printk(KERN_NOTICE PFX "Size for event buff is less"
			" then %d, set default value (%d)\n",
			GALIL1800_EVENT_BUFF_SZ_MIN, *ev_buff_sz);
		return 1;
	}
	else if(*ev_buff_sz > GALIL1800_EVENT_BUFF_SZ_MAX) {
		/* A circular buffer cannot work!! */
		*ev_buff_sz = GALIL1800_EVENT_BUFF_DEFAULT_SZ;
		printk(KERN_NOTICE PFX "Size for event buff is more"
			" then %d, so set to default value (%d)\n",
			GALIL1800_EVENT_BUFF_SZ_MAX, *ev_buff_sz);
		return 2;
	}
	else if(*ev_buff_sz % GALIL1800_EVENT_BUFF_FRAME_SZ) {
		*ev_buff_sz -= *ev_buff_sz % GALIL1800_EVENT_BUFF_FRAME_SZ;
		printk(KERN_WARNING PFX "Size for event buff must be "
				"multiple of %d, so truncated to %d.\n",
				GALIL1800_EVENT_BUFF_FRAME_SZ, *ev_buff_sz);
		return 3;
	}
	return 0;
}

/* Second step of initialization: must be call in open device function */
static int init_galil1800_event_buffer_open_dev(
			struct galil1800_event_buffer *buff_p)
{
	/* Not need to protect pointer whit lock, because if otp_flag
	is 0 and event device file is closed, noboby use them.
	Note that this function needs to be call whit open semaphore down!! */
	int i;
	/* event buffer size more another buffer frame for special purpose */
	size_t buffer_size = sizeof(struct galil1800_event_data) *
				(buff_p->size+GALIL1800_EVENT_BUFF_FRAME_SZ);

	buff_p->frames = kzalloc(buffer_size, GFP_KERNEL);
	if (!(buff_p->frames)) {
		printk(KERN_ERR PFX "init_galil1800_event_buffer_open_dev:"
						" alloc frames failed\n");
		return -ENOMEM;
	}

	buff_p->table = kcalloc(GALIL1800_EVENT_BUFF_TABLE_SZ(buff_p),
				sizeof(struct galil1800_event_data*),
				GFP_KERNEL);
	if (!(buff_p->table)) {
		printk(KERN_ERR PFX "init_galil1800_event_buffer_open_dev"
				": alloc table failed\n");
		kfree(buff_p->frames);
		return -ENOMEM;
	}

	for(i=0; i < GALIL1800_EVENT_BUFF_TABLE_SZ(buff_p); i++) {
		buff_p->table[i] = buff_p->frames +
				(i * GALIL1800_EVENT_BUFF_FRAME_SZ);
	}
	buff_p->read_frame = buff_p->frames +
				(i * GALIL1800_EVENT_BUFF_FRAME_SZ);

	/* Set default opt_flag */
	galil1800_event_active_default(buff_p);
	return 0;
}

/* Clean for second step of initialization:
   must be call in release device function */
static inline void clean_galil1800_event_buffer_close_dev(
				struct galil1800_event_buffer *buff_p)
{
	spin_lock_bh(&buff_p->lock_pt);
	/* Here I must acquire spinlock because an add function
	could be use the buffer, while I free it! */
	galil1800_event_disactive_all(buff_p);
	spin_unlock_bh(&buff_p->lock_pt);

	/* Not need to protect pointer whit lock, because if otp_flag
	is 0 and event device file is closed, noboby use them.
	Note that this function needs to be call whit open semaphore down!! */
	kfree(buff_p->frames);
	kfree(buff_p->table);

	buff_p->first_frame = 0;
	buff_p->read_frame = 0;
	buff_p->written = 0;
	buff_p->overwritten_frames = 0;
}

static void galil1800_event_buff_prepare_to_read(
		struct galil1800_event_buffer *buff_p,
		unsigned int *temp_overw_f, unsigned int *temp_written)
{
	struct galil1800_event_data *temp;

	spin_lock_bh(&buff_p->lock_pt);

	*temp_overw_f = buff_p->overwritten_frames;
	*temp_written = buff_p->written;

	if(buff_p->written > 0) {
		buff_p->overwritten_frames = 0;
		temp = buff_p->read_frame;
		barrier();
		buff_p->read_frame = buff_p->table[buff_p->first_frame];
		barrier();
		buff_p->table[buff_p->first_frame] = temp;
		if(buff_p->written >= GALIL1800_EVENT_BUFF_FRAME_SZ) {
			buff_p->written -= GALIL1800_EVENT_BUFF_FRAME_SZ;
			galil1800_event_buff_first_frame_inc(buff_p);
		} else
			buff_p->written -= buff_p->written %
				GALIL1800_EVENT_BUFF_FRAME_SZ;
	}

	spin_unlock_bh(&buff_p->lock_pt);
}

static ssize_t galil1800_event_read(struct file * filp,
		char __user * buf, size_t size, loff_t *poff)
{
	struct galil1800_state *s =
		(struct galil1800_state *)filp->private_data;
	struct galil1800_event_buffer *buff_p = &s->event_buff;
	const size_t min_data_size =
		sizeof(struct galil1800_event_data) *
		(GALIL1800_EVENT_BUFF_FRAME_SZ + 1);
	unsigned int rc=0, lost_events, written_events, bytes;
	ssize_t request = size;
	unsigned char *_buf;

	GALIL_PRINT_DEBUG_PID("\"event read\" called\n");

	if (!buf)
		return -EFAULT;

	if (request < min_data_size) {
		printk(KERN_NOTICE PFX "event read: size is less then min."
			" data size (%d)\n", min_data_size);
		return -EFAULT;
	}

	request -= (request % sizeof(struct galil1800_event_data));

	/* I want to limit the maximun lenght of temp. _buf */
	request = (request <
		(GALIL1800_EVENT_BUFF_TABLE_SZ(buff_p)*min_data_size) ?
		request :
		(GALIL1800_EVENT_BUFF_TABLE_SZ(buff_p)*min_data_size));

	if (!(_buf = kzalloc(request, GFP_KERNEL))) {
		printk(KERN_ERR PFX "galil1800_event_read:"
			" alloc of _buf failed\n");
		return -ENOMEM;
	}

	/* Wait to data arrive */
	while(galil1800_event_buff_empty(buff_p)) {
		if(filp->f_flags & O_NONBLOCK) {
			rc = -EAGAIN;
			goto err;
		}
		if(wait_event_interruptible(buff_p->read_wait,
				!galil1800_event_buff_empty(buff_p))) {
			rc = -ERESTARTSYS;
			goto err;
		}
	}

	do {
		galil1800_event_buff_prepare_to_read(buff_p,
				&lost_events, &written_events);
		if(written_events > 0) {
		/* now I can write max (GALIL1800_EVENT_BUFF_FRAME_SZ + 1) *
		   sizeof(struct galil1800_event_data) */
			if(lost_events > 0) {
				lost_events *= GALIL1800_EVENT_BUFF_FRAME_SZ;
				_buf[rc++] = GALIL1800_EVENT_BUFF_OVERWRITTEN;
				_buf[rc++] = (unsigned char)(lost_events < 256 ?
						lost_events : 0);
			}
			bytes = written_events < GALIL1800_EVENT_BUFF_FRAME_SZ ?
				written_events % GALIL1800_EVENT_BUFF_FRAME_SZ :
				GALIL1800_EVENT_BUFF_FRAME_SZ;
			bytes*=sizeof(struct galil1800_event_data);
			memcpy(_buf + rc, buff_p->read_frame, bytes);
			rc+=bytes;
		}
		else
			break;
	}while( (rc + (GALIL1800_EVENT_BUFF_FRAME_SZ + 1) *
		 sizeof(struct galil1800_event_data)) <= request );

	if(rc>0) {
		if(copy_to_user(buf, _buf, rc))
		rc = -EFAULT;
	}
err:
	kfree(_buf);
	return rc;
}

static ssize_t galil1800_event_write(struct file * filp,
		const char __user * buf, size_t size, loff_t *poff)
{
	return -EPERM;
}

static unsigned int galil1800_event_poll(struct file *filp, poll_table *wait)
{
	struct galil1800_state *s =
		(struct galil1800_state *) filp->private_data;
	unsigned int mask = 0;
	poll_wait(filp, &(s->event_buff.read_wait), wait);
	if(!galil1800_event_buff_empty(&(s->event_buff)))
		mask = POLLIN | POLLRDNORM;
	return mask;
}

static int galil1800_event_open(struct inode * inode, struct file * filp)
{
	struct galil1800_state *s;
	struct galil1800_devfile_state *df_s;
	int ret;

	s = container_of(inode->i_cdev, struct galil1800_state,
			devfile_state[GALIL1800_EVENT_DEVICE].cdev);
	filp->private_data = s;
	df_s = &(s->devfile_state[GALIL1800_EVENT_DEVICE]);

	/* wait for device to become free */

	GALIL_PRINT_DEBUG_PID("\"event open\" started\n");

	if(down_interruptible(&df_s->open_sem)) {
		GALIL_PRINT_DEBUG_PID("\"event open\": \"down\" interrupted\n");
		return -ERESTARTSYS;
	}

	/* Whit this type of check,
	   one process can open for read and another for write */

	if(filp->f_mode & FMODE_WRITE) {
		GALIL_PRINT_DEBUG_PID("\"event open\" failed: FMODE_WRITE"
				      " is set!\n");
		up(&df_s->open_sem);
		return -EACCES;
	}

	while (df_s->open_mode) {
		if (filp->f_flags & O_NONBLOCK) { /* O_NONBLOCK == O_NDELAY */
			up(&df_s->open_sem);
			GALIL_PRINT_DEBUG_PID("\"event open\" return:"
				" O_NONBLOCK is set and device is busy\n");
			return -EAGAIN;
		}
		up(&df_s->open_sem);
		NO_GALIL_ONE_PRINT_DEBUG_PID("\"event open\" going to sleep\n");
		if(wait_event_interruptible(df_s->open_wait,df_s->open_mode==0)) {
			GALIL_PRINT_DEBUG_PID("\"event open\":"
					" \"wait_event\" interrupted\n");
			return -ERESTARTSYS;
		}
		if(down_interruptible(&df_s->open_sem)) {
			GALIL_PRINT_DEBUG_PID("\"event open\":"
				" \"down\" interrupted\n");
			return -ERESTARTSYS;
		}
	}

	ret = init_galil1800_event_buffer_open_dev(&s->event_buff);
	if(ret < 0) {
		up(&df_s->open_sem);
		return ret;
	}

	ret = galil1800_init_controller_open_step(s);
	if(ret < 0) {
		clean_galil1800_event_buffer_close_dev(&s->event_buff);
		up(&df_s->open_sem);
		return ret;
	}

	df_s->open_mode = filp->f_mode & FMODE_READ;

	up(&df_s->open_sem);
	nonseekable_open(inode,filp);
	GALIL_PRINT_DEBUG_PID("\"event open\" ended\n");
	return 0;
}

static int galil1800_event_release(struct inode *inode, struct file *filp)
{
	struct galil1800_state *s =
		(struct galil1800_state *) filp->private_data;
	struct galil1800_devfile_state *df_s =
		&(s->devfile_state[GALIL1800_EVENT_DEVICE]);

	GALIL_PRINT_DEBUG_PID("\"event release\" started\n");
	if(down_interruptible(&df_s->open_sem)) {
		GALIL_PRINT_DEBUG_PID("\"event release\":"
			" \"down\" interrupted\n");
		return -ERESTARTSYS;
	}

	df_s->open_mode &= (~filp->f_mode) & FMODE_READ;

	clean_galil1800_event_buffer_close_dev(&s->event_buff);

	galil1800_clean_controller_open_step(s);

	up(&df_s->open_sem);
	wake_up_interruptible_sync(&df_s->open_wait);
	GALIL_PRINT_DEBUG_PID("\"event release\" ended:"
			" wake up of sleeping open\n");
	return 0;
}

static long galil1800_event_unlocked_ioctl(struct file *filp,
				unsigned int cmd, unsigned long arg)
{
	struct galil1800_state *dev =
		(struct galil1800_state *) filp->private_data;
	struct galil1800_event_buffer *buff_p = &dev->event_buff;

	/* Check command */
	if(_IOC_TYPE(cmd) != GALIL1800_IOCTL_MAGIC)
		return -ENOTTY;
	if(_IOC_NR(cmd) < GALIL1800_IOCTL_EVENT_MINNR ||
		_IOC_NR(cmd) > GALIL1800_IOCTL_EVENT_MAXNR)
		return -ENOTTY;

	/* Check argument:
		After this check, we safely use:
		__copy_to_user or __copy_from_user
		__put_user or __get_user          */
	if(_IOC_DIR(cmd) & _IOC_READ) {
		/* Read request, so check if I can write to user space */
		if(!(access_ok(VERIFY_WRITE,(void __user *)arg,_IOC_SIZE(cmd))))
			return -EFAULT;
	}
	else if(_IOC_DIR(cmd) & _IOC_WRITE) {
		/* Write request, so check if I can read from user space */
		if(!(access_ok(VERIFY_READ,(void __user *)arg,_IOC_SIZE(cmd))))
			return -EFAULT;
	}

	switch (cmd) {
		case GALIL1800_IOCTL_EVENT_ALL_ACTIVE:
		{
			if((void*)arg==NULL) /* disable */
				galil1800_event_disactive_all(buff_p);
			else /* enable */
				galil1800_event_active_all(buff_p);
		}
		break;
		case GALIL1800_IOCTL_EVENT_IRQ:
		{
			if((void*)arg==NULL) /* disable */
				galil1800_event_disactive(buff_p,
						GALIL1800_EVENT_IRQ);
			else /* enable */
				galil1800_event_active(buff_p,
						GALIL1800_EVENT_IRQ);
		}
		break;
		case GALIL1800_IOCTL_EVENT_WBUFF_LESS_10P:
		{
			if((void*)arg==NULL) /* disable */
				galil1800_event_disactive(buff_p,
					GALIL1800_EVENT_WBUFF_LESS_10P);
			else /* enable */
				galil1800_event_active(buff_p,
					GALIL1800_EVENT_WBUFF_LESS_10P);
		}
		break;
		case GALIL1800_IOCTL_EVENT_WBUFF_LESS_20P:
		{
			if((void*)arg==NULL) /* disable */
				galil1800_event_disactive(buff_p,
					GALIL1800_EVENT_WBUFF_LESS_20P);
			else /* enable */
				galil1800_event_active(buff_p,
					GALIL1800_EVENT_WBUFF_LESS_20P);
		}
		break;
		case GALIL1800_IOCTL_EVENT_WBUFF_LESS_50P:
		{
			if((void*)arg==NULL) /* disable */
				galil1800_event_disactive(buff_p,
					GALIL1800_EVENT_WBUFF_LESS_50P);
			else /* enable */
				galil1800_event_active(buff_p,
					GALIL1800_EVENT_WBUFF_LESS_50P);
		}
		break;
		case GALIL1800_IOCTL_EVENT_WBUFF_OVER_90P:
		{
			if((void*)arg==NULL) /* disable */
				galil1800_event_disactive(buff_p,
					GALIL1800_EVENT_WBUFF_OVER_90P);
			else /* enable */
				galil1800_event_active(buff_p,
					GALIL1800_EVENT_WBUFF_OVER_90P);
		}
		break;
		default:
			return -ENOTTY; /* cmd not valid */
	}
	return 0; /* Success */
}


struct file_operations galil1800_event_fops = {
	.owner = THIS_MODULE,
	.read = galil1800_event_read,
	.write = galil1800_event_write,
	.poll = galil1800_event_poll,
	.unlocked_ioctl = galil1800_event_unlocked_ioctl,
	.open = galil1800_event_open,
	.release = galil1800_event_release,
	.llseek = no_llseek
};

static
ssize_t galil1800_sysfs_event_buff_sz_show(struct class_device *cd, char *buf)
{
	struct galil1800_devfile_state *df_s =
		(struct galil1800_devfile_state *)(cd->class_data);

	struct galil1800_state *s = container_of(df_s,
				struct galil1800_state,
				devfile_state[GALIL1800_EVENT_DEVICE]);

	return sprintf(buf, "%u\n", s->event_buff.size);
}

static
ssize_t galil1800_sysfs_event_buff_sz_store(struct class_device *cd,
						const char *buf, size_t count)
{
	struct galil1800_devfile_state *df_s =
		(struct galil1800_devfile_state *)(cd->class_data);
	struct galil1800_state *s = container_of(df_s,
				struct galil1800_state,
				devfile_state[GALIL1800_EVENT_DEVICE]);

	GALIL_PRINT_DEBUG("galil1800_sysfs_event_buff_sz_store, buf: \"%s\"\n",
									buf);

	if(down_interruptible(&df_s->open_sem)) {
		GALIL_PRINT_DEBUG_PID("\"event_buff_sz_store\": \"down\""
					" interrupted\n");
		return -ERESTARTSYS;
	}

	if(df_s->open_mode) {
		up(&df_s->open_sem);
		printk(KERN_ERR PFX "event_buff_sz not modified."
					" There are files opened\n");
		return -EBUSY;
	}
	if(count>10 || strlen(buf)!=count) {
		up(&df_s->open_sem);
		printk(KERN_ERR PFX "event_buff_sz not modified."
					" Too much characters\n");
		return -EINVAL;
	}

	sscanf(buf, "%u", &s->event_buff.size);

	galil1800_event_buff_size_check(&s->event_buff.size);

	up(&df_s->open_sem);

	return count;
}

/* Add event cdev device and register it in sysfs */
int galil1800_event_cdev_add(struct cdev *cdev_p,
			unsigned int major, unsigned int minor)
{
	int ret;
	struct galil1800_state *s = container_of(cdev_p,
				struct galil1800_state,
				devfile_state[GALIL1800_EVENT_DEVICE].cdev);

	cdev_init(cdev_p, &galil1800_event_fops);

	cdev_p->owner = THIS_MODULE;

	ret=cdev_add(cdev_p, MKDEV(major, minor), 1);
	if (ret) {
		printk(KERN_INFO PFX "cannot add event cdev dev=(%d,%d)\n",
							major, minor);
		return ret;
	}

	ret=galil1800_sysfs_class_dev_register(s, GALIL1800_EVENT_DEVICE);
	if(ret) {
		printk(KERN_INFO PFX "sysfs class device register"
					" failed for event device\n");
		goto err_class_dev_reg;
	}

	s->event_buff.cd_attr_size.attr.name = "event_buff_sz";
	s->event_buff.cd_attr_size.attr.owner = THIS_MODULE;
	s->event_buff.cd_attr_size.attr.mode = S_IWUSR | S_IRUGO;
	s->event_buff.cd_attr_size.show  = galil1800_sysfs_event_buff_sz_show;
	s->event_buff.cd_attr_size.store = galil1800_sysfs_event_buff_sz_store;
	ret=class_device_create_file(
		&s->devfile_state[GALIL1800_EVENT_DEVICE].class_dev,
		&s->event_buff.cd_attr_size);
	if(ret) {
		printk(KERN_ERR PFX "cannot create file for class "
			"device attr \"%s\" in sysfs\n",
			s->event_buff.cd_attr_size.attr.name);
		goto err_attr;
	}

	return 0;

err_attr:
	galil1800_sysfs_class_dev_unregister(s, GALIL1800_EVENT_DEVICE);
err_class_dev_reg:
	cdev_del(cdev_p);
	return ret;
}

/* Delete event cdev device and deregister it in sysfs */
void galil1800_event_cdev_del(struct cdev *cdev_p)
{
	struct galil1800_state *s = container_of(cdev_p,
				struct galil1800_state,
				devfile_state[GALIL1800_EVENT_DEVICE].cdev);

	galil1800_sysfs_class_dev_unregister(s, GALIL1800_EVENT_DEVICE);

	class_device_remove_file(
		&s->devfile_state[GALIL1800_EVENT_DEVICE].class_dev,
		&s->event_buff.cd_attr_size );

	cdev_del(cdev_p);
}
