#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>

#include "galil1800.h"

#define GALIL_CARD_DEV   "/dev/galil/card0"
#define GALIL_EVENT_DEV  "/dev/galil/event0"
#define GALIL_INFO_DEV   "/dev/galil/info0"
#define SERIAL_IN_SYSFS  "/sys/class/galil/card0/device/serial"

int test1(int);
int test2(int, int);
int main(void);
void init(int *, int *, pid_t *);
void deinit(int, int, pid_t *);
void read_event(void);
int send(int, char *);
void read_answer(void);

#define SEND(_x, _y)              \
	do {                      \
		if(send(_x, _y))  \
			return 1; \
	}while(0)


int test0(int c) {
	int fd, rc;
	char serial[GALIL1800_SERIAL_MAXLEN+1];

	printf("\n    [ ---- START TEST 0 ---- ]\n");

	printf("\nSerial number can be read from sysfs...\n");

	fd = open(SERIAL_IN_SYSFS, O_RDONLY);
	if (fd < 0) {
		perror("open " SERIAL_IN_SYSFS);
		return 1;
	}
	rc = read(fd, serial, GALIL1800_SERIAL_MAXLEN);
	if (rc <= 0) {
		perror("read " SERIAL_IN_SYSFS);
		return 1;
	}
	close(fd);
	serial[rc]=0;
	printf("\n\t Serial number read from %s: %s\n",
				SERIAL_IN_SYSFS, serial);

	printf("\nOr you can get it with an ioctl command...\n");

	memset(serial, 0, GALIL1800_SERIAL_MAXLEN+1);

	rc = ioctl(c, GALIL1800_IOCTL_GET_SERIAL_NUMBER, serial);
	if (rc) {
		perror("ioctl for \"serial number\"");
		return 1;
	}
	printf("\n\t Serial number got with ioctl command: %s\n", serial);

	return 0;
}


int test1(int c)
{
	printf("\n\n    [ ---- START TEST 1 ---- ]\n\n");

	SEND(c, "EO1;");
	SEND(c, "MG \"Test #1\";");
	SEND(c, ";SH;");
	SEND(c, "LA;LL;LV;");
	SEND(c, "LS0,999;");
	SEND(c, "TB?;");
	printf("QZ command output:\n\t(1) number of axes,\n"
		"\t(2) number of bytes to be transferred for general status,\n"
		"\t(3) number bytes to be transferred for coordinated"
							" move status,\n"
	       "\t(4) number of bytes to be transferred for axis specific"
							" information\n");
	SEND(c, "QZ;");
	SEND(c, "RP;");
	SEND(c, "SC;");
	SEND(c, "TC;");
	SEND(c, "no;TC;");
	printf("When using buffering, the application may need to unambiguously"
		"recognize\nthe data coming from the card. A simple method"
		"consists of marking each group of\ndata with the MG"
		" command\n");
	SEND(c, "MG \"TD\"{N};TD;");
	SEND(c, "MG \"TE\"{N};TE;");
	SEND(c, "TI;");
	printf("From the manual: \"Note that a value of 1000 for the update"
		" rate (TM command)\n"
		"will actually set an update rate of 1024 microseconds."
		" Thus the value returned\nby the TIME operand will be off by"
		" 2.4%% of the actual time.\"\n");
	SEND(c, "MG TIME,\",\",_TM;");
	SEND(c, "TS;");
	SEND(c, "TT;");
	SEND(c, "TV;");
	SEND(c, "MG @IN[0],@OUT[0],@AN[0];");
	SEND(c, "MG @RND[@SQR[9.7]];");
	SEND(c, "MG @SQR[-9.0];");
	SEND(c, "MG \"BEFORE WT:\",TIME;");
	SEND(c, "WT 2000;");
	SEND(c, "MG \"AFTER WT:\",TIME;");
	SEND(c, "EO0;");
	printf("Application sleeping 5 seconds\n");
	fflush(stdout);
	sleep(5);
	SEND(c, ";");
	SEND(c, "UI0;UI1;UI2;UI3;WT1000;");
	SEND(c, "UI4;UI5;UI6;UI7;WT1000;");
	SEND(c, "UI8;UI9;UI10;UI11;WT1000;");
	SEND(c, "UI12;UI13;UI14;UI15;");
	SEND(c, "WT1000;");
	printf("Application sleeping 3 second\n");
	sleep(3);
	SEND(c, ";");

	return 0;
}

int test2(int c, int i) {
	unsigned char buf[GALIL1800_MAX_FIFO2_SIZE];
	unsigned char *p = buf;
	int n, rc;

	printf("\n\n    [ ---- START TEST 2 ---- ]\n\n");

	SEND(c, "DR1;");

	for(n=0; n<3; n++) {
		sleep(1);
		rc = read(i, p , GALIL1800_MAX_FIFO2_SIZE);
		if (rc == -1) {
			perror("read FIFO2");
			return 1;
		}

		printf("[INFO] #%d: Read %d bytes from secondary FIFO\n",
								n, rc);

		if(rc) {
			int n;
			printf("[INFO] Secondary FIFO dump:\n");
			for(n=0; n<rc; n++) {
				if(!(n%16))
					printf("\n");
				printf(" %02x", buf[n]);
			}
			printf("\n\n");
		}
	}

	SEND(c, "DR0;");

	return 0;
}

int main()
{
	pid_t pids[2];
	int fd_card, fd_info;

	init(&fd_card, &fd_info, pids);

	if(test0(fd_card))
		goto out;
	sleep(1);

	if(test1(fd_card))
		goto out;

	sleep(2);

	if(test2(fd_card, fd_info))
		goto out;

out:
	deinit(fd_card, fd_info, pids);
	return 0;
}

void init(int *pc, int *pi, pid_t *pids)
{
	int c, i;

	pids[0] = fork();
	if (pids[0] == -1) {
		perror("fork event");
		exit(1);
	}
	if (!pids[0])
		read_event();
	else {
		pids[1] = fork();
		if (pids[1] == -1) {
			perror("fork info");
			kill(pids[0], SIGKILL);
			exit(1);
		}
		if (!pids[1])
			read_answer();
		else {
			c = open(GALIL_CARD_DEV, O_WRONLY);
			i = open(GALIL_INFO_DEV, O_RDONLY);
			if (c == -1 || i == -1) {
				perror("open for " GALIL_CARD_DEV
					" or " GALIL_INFO_DEV);
				kill(pids[0], SIGKILL);
				kill(pids[1], SIGKILL);
				exit(1);
			}
			*pc = c;
			*pi = i;
		}
	}
}

void deinit(int c, int i, pid_t *pids)
{
	fflush(stdout);
	sleep(1);
	if (close(c) == -1)
		perror("close " GALIL_CARD_DEV);
	if (close(i) == -1)
		perror("close " GALIL_INFO_DEV);
	kill(pids[0], SIGKILL);
	kill(pids[1], SIGKILL);
}

char * eventdump(char *buf, int len)
{
	static char dump[255*7];
	char *p;
	int i;
	const int size = sizeof(struct galil1800_event_data);

	for (i=0, p=dump; i<len; i+=size) {
		struct galil1800_event_data *ev =
			(struct galil1800_event_data*) (buf+i);
		p+=sprintf(p, "{%02x,%02x}", ev->type, ev->data);
	}
	*p = '\0'; /* len can be 0 */
	return (char *)dump;
}

void read_event(void)
{
	int e;
	char buf[513];
	int len;

	e = open(GALIL_EVENT_DEV, O_RDONLY);
	if (e == -1) {
		perror("open " GALIL_EVENT_DEV);
		exit(1);
	}
	for (;;) {
		len = read(e, &buf, 512);
		if(len == -1) {
			perror("read " GALIL_EVENT_DEV);
			exit(1);
		}
		if(!len)
			continue;

		buf[len] = '\0';
		fprintf(stderr, "EVENT: (%d) %s\n", len, eventdump(buf,len));
	}
}

int send(int f, char *cmd)
{
	int l, lc, rc;

	lc = strlen(cmd);
	for (l=0; l<lc; l+=rc) {
		rc = write(f, cmd+l , lc-l);
		if (rc == -1) {
			perror("write commands");
			return 1;
		}
	}

	printf("CARD: written '%s'\n", cmd);

	return 0;
}

void read_answer(void) {
	char buf[GALIL1800_READ_FIFO_SIZE+1];
	int fd, rc;

	fd = open(GALIL_CARD_DEV, O_RDONLY);
	if (fd == -1) {
		perror("read only open for " GALIL_CARD_DEV);
		exit(1);
	}

	while(1) {
		rc = read(fd, buf, GALIL1800_READ_FIFO_SIZE);
		if (rc == -1) {
			perror("CARD: [error while reading answer]\n");
			exit(1);
		}

		if (rc == 0) { /* Only for op. with O_NONBLOCK  */
			printf("CARD: [no answer]\n");
			continue;
		}

		printf("\n\nCARD: [read %d bytes}:\n", rc);
		buf[rc]=0;
		printf("%s\n", buf);
	}
}
